//
//  AppDelegate.m
//  SocketServer
//
//  Created by Clemens Wagner on 05.09.15.
//  Copyright © 2015 Clemens Wagner. All rights reserved.
//

#import "AppDelegate.h"
#import "Connection.h"
#import "FMDB.h"
#import "NSURL+Extensions.h"

static NSString * kServerDomain = @"local.";
static NSString * kServerName = @"LocalHTTPServer";
// Besser: Dynamischer Port (= 0) und über server.port abfragen
static const NSInteger kServerPort = 8888;
static NSString * kServerBonjourType = @"_http._tcp.";

@interface AppDelegate ()<NSNetServiceDelegate, ConnectionDelegate>

@property (strong, nonatomic) NSNetService *server;
@property (strong, nonatomic) NSMutableArray *connections;
@property (strong, nonatomic) FMDatabaseQueue *databaseQueue;
@property (copy, nonatomic) NSDictionary *mimeTypes;

@end

@implementation AppDelegate


- (BOOL)application:(UIApplication *)application didFinishLaunchingWithOptions:(NSDictionary *)launchOptions {
    NSString *thePath = [[NSBundle mainBundle] pathForResource:@"data" ofType:@"db"];
    
    self.connections = [NSMutableArray array];
    self.mimeTypes = @{
                       @"css": @"text/css",
                       @"html": @"text/html; charset: UTF-8",
                       @"js": @"text/javascript"
                       };
    self.databaseQueue = [FMDatabaseQueue databaseQueueWithPath:thePath flags:SQLITE_OPEN_READONLY];
    [self startSocketServer];
    return YES;
}

- (void)startSocketServer {
    NSNetService *theServer = [[NSNetService alloc] initWithDomain:kServerDomain
                                                              type:kServerBonjourType
                                                              name:kServerName port:kServerPort];
    theServer.includesPeerToPeer = NO;
    [theServer setDelegate:self];
    [theServer publishWithOptions:NSNetServiceListenForConnections];
    self.server = theServer;
}

- (NSString *)mimeTypeForFile:(NSString *)inFile {
    NSString *theExtension = [[inFile pathExtension] lowercaseString];
    NSString *theType = self.mimeTypes[theExtension];
    
    return theType == nil ? @"application/octet-stream" : theType;
}

- (NSCachedURLResponse *)responseForFile:(NSString *)inFile withURL:(NSURL *)inURL {
    NSURL *theURL = [[NSBundle mainBundle] URLForResource:inFile withExtension:nil];
    NSData *theData = [NSData dataWithContentsOfURL:theURL];
    NSHTTPURLResponse *theResponse;
    
    if(theData == nil) {
        theResponse = [[NSHTTPURLResponse alloc] initWithURL:inURL
                                                  statusCode:404
                                                 HTTPVersion:@"HTTP/1.1"
                                                headerFields:@{}];
        theData = [NSData data];
    }
    else {
        NSString *theContentType = [self mimeTypeForFile:inFile];
        NSDictionary *theHeaders = @{
                                     @"Content-Type": theContentType,
                                     @"Content-Length": [NSString stringWithFormat:@"%ld", (long)theData.length],
                                     };
        
        theResponse = [[NSHTTPURLResponse alloc] initWithURL:inURL
                                                  statusCode:200
                                                 HTTPVersion:@"HTTP/1.1"
                                                headerFields:theHeaders];
    }
    return [[NSCachedURLResponse alloc] initWithResponse:theResponse data:theData];
}

- (NSCachedURLResponse *)responseWithJSONData:(id)inData forURL:(NSURL *)inURL {
    NSError *theError = nil;
    NSData *theData = inData == nil || [inData isKindOfClass:[NSData class]] ? inData :
    [NSJSONSerialization dataWithJSONObject:inData options:0 error:&theError];
    NSHTTPURLResponse *theResponse;
    
    if(theData == nil) {
        NSLog(@"error = %@", theError);
        theResponse = [[NSHTTPURLResponse alloc] initWithURL:inURL
                                                  statusCode:404
                                                 HTTPVersion:@"HTTP/1.1"
                                                headerFields:@{}];
        theData = [NSData data];
    }
    else {
        NSDictionary *theHeaders = @{
                                     @"Content-Type": @"application/json",
                                     @"Content-Length": [NSString stringWithFormat:@"%ld", (long)theData.length],
                                     };
        
        theResponse = [[NSHTTPURLResponse alloc] initWithURL:inURL
                                                  statusCode:200
                                                 HTTPVersion:@"HTTP/1.1"
                                                headerFields:theHeaders];
    }
    return [[NSCachedURLResponse alloc] initWithResponse:theResponse data:theData];
}

#pragma mark NSNetServiceDelegate

- (void)netService:(NSNetService *)inService didAcceptConnectionWithInputStream:(NSInputStream *)inInputStream outputStream:(NSOutputStream *)inOutputStream {
    Connection *theConnection = [[Connection alloc] initWithName:inService.name inputStream:inInputStream outputStream:inOutputStream];
    
    theConnection.delegate = self;
    @synchronized(self) {
        [self.connections addObject:theConnection];
    }
}

#pragma mark ConnectionDelegate

- (NSCachedURLResponse *)connection:(Connection *)inConnection responseForRequest:(NSURLRequest *)inRequest {
    NSURL *theURL = inRequest.URL;
    NSString *thePath = theURL.path;
    
    NSLog(@"url = %@", theURL);
    if([thePath hasPrefix:@"/autocompletion"]) {
        NSString *theTerm = [theURL.parameters[@"term"] lowercaseString];
        NSUInteger theLength = theTerm.length;
        NSMutableArray *theResult = nil;
        
        if(theLength > 0) {
            NSString *theUpperTerm = [NSString stringWithFormat:@"%@%C",
                                      [theTerm substringToIndex:theLength - 1],
                                      (unichar)([theTerm characterAtIndex:theLength - 1] + 1)];
            theResult = [NSMutableArray new];
            [self.databaseQueue inDatabase:^(FMDatabase *inDatabase) {
                FMResultSet *theSet = [inDatabase executeQuery:@"SELECT title FROM lexicon WHERE key >= ? AND key < ? ORDER BY key LIMIT 20",
                                       theTerm, theUpperTerm];
                
                while (theSet.next) {
                    NSString *theTitle = [theSet stringForColumnIndex:0];
                    
                    [theResult addObject:theTitle];
                }
                [theSet close];
            }];
        }
        return [self responseWithJSONData:theResult forURL:theURL];
    }
    else if([thePath hasPrefix:@"/search"]) {
        NSString *theTerm = theURL.parameters[@"term"];
        __block NSData *theData = nil;

        [self.databaseQueue inDatabase:^(FMDatabase *inDatabase) {
            FMResultSet *theSet = [inDatabase executeQuery:@"SELECT data FROM lexicon WHERE title =? LIMIT 1", theTerm];
            
            if(theSet.next) {
                NSString *theText = [theSet stringForColumnIndex:0];
                
                theData = [theText dataUsingEncoding:NSUTF8StringEncoding];
            }
            [theSet close];
        }];
        return [self responseWithJSONData:theData forURL:theURL];
    }
    else {
        if([thePath isEqualToString:@"/"]) {
            thePath = @"index.html";
        }
        return [self responseForFile:[thePath lastPathComponent] withURL:theURL];
    }
}

- (void)connection:(Connection *)inConnection didWriteResponse:(NSCachedURLResponse *)inResponse {
    @synchronized(self) {
        [self.connections removeObject:inConnection];
    }
}
@end
